# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import re
from bs4 import BeautifulSoup
from school_api.exceptions import OtherException


def get_alert_tip(html):
    ''' 获取页面提示信息 '''
    tips = re.findall(r">alert\(\'(.*?)\'", html)
    if tips:
        return tips[0]


def get_view_state_from_html(html):
    ''' 获取 __VIEWSTATE 值 '''
    pre_soup = BeautifulSoup(html, "html.parser")
    view_state_soup = pre_soup.find(attrs={"name": "__VIEWSTATE"})
    try:
        view_state = view_state_soup['value']
    except TypeError:
        raise OtherException('', '获取view_state失败')

    return view_state



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from school_api.utils import to_text, ObjectDict
from school_api.config import URL_ENDPOINT, CLASS_TIME
from school_api.client.base import BaseUserClient
from school_api.client.api.score import Score
from school_api.client.api.schedule import Schedule
from school_api.client.api.place_schedule import PlaceSchedule
from school_api.client.api.user_info import UserlInfo
from school_api.client.utils import UserType, error_handle, ApiPermissions, get_time_list
from school_api.session.memorystorage import MemoryStorage


class SchoolClient(object):
    ''' 学校实例 '''

    def __init__(self, url, name=None, code=None, use_ex_handle=True, exist_verify=True, lan_url=None, proxies=None,
                 priority_proxy=False, timeout=10, login_url_path='/default2.aspx', url_endpoint=URL_ENDPOINT,
                 class_time_list=CLASS_TIME, session=MemoryStorage):
        self.base_url = url.split('/default')[0] if url[-4:] == 'aspx' else url
        self.school = {
            'name': to_text(name),
            'code': code,
            'use_ex_handle': use_ex_handle,
            'exist_verify': exist_verify,
            'lan_url': lan_url,
            'proxies': proxies,
            'priority_proxy': priority_proxy,
            'timeout': timeout,
            'login_url': login_url_path,
            'url_endpoint': url_endpoint,
            'time_list': get_time_list(class_time_list)
        }
        self.session = session(self.school['code'])
        self.school = ObjectDict(self.school)

    def user_login(self, account, password, **kwargs):
        ''' 用户登录入口
        进行首次绑定操作时，请将 use_login_cookie 设置为False，避免其他用户进行会话登录
        :param account:  用户账号
        :param password: 用户密码
        :param user_type: 0.学生 1.教师 2.部门
        :param use_login_cookie: 是否使用会话登陆
        :param requests模块参数
        return 用户实例
        '''
        use_login_cookie = not (not kwargs.pop('use_session', True) or not kwargs.pop('use_login_cookie', True))
        user_type = kwargs.pop('user_type', UserType.STUDENT)
        user = UserClient(self, account, password, user_type)
        user = user.user_login(use_login_cookie, **kwargs)
        return user


class UserClient(BaseUserClient):
    ''' 用户实例 '''

    score = Score()
    info = UserlInfo()
    schedule = Schedule()
    place_schedule = PlaceSchedule()

    @error_handle
    def user_login(self, use_login_cookie, **kwargs):
        ''' 登录：通过SchoolClient类调用 '''
        login_session = None
        if use_login_cookie:
            login_session = self.session_management()

        if login_session is None:
            self.login.get_login(self.school, **kwargs)
            self.save_login_session()
            login_session = self

        return login_session

    @error_handle
    def get_schedule(self, *args, **kwargs):
        return self.schedule.get_schedule(*args, **kwargs)

    @error_handle
    @ApiPermissions([UserType.STUDENT, UserType.TEACHER])
    def get_info(self, **kwargs):
        return self.info.get_info(**kwargs)

    @error_handle
    @ApiPermissions([UserType.STUDENT])
    def get_score(self, *args, **kwargs):
        return self.score.get_score(*args, **kwargs)

    @error_handle
    @ApiPermissions([UserType.DEPT])
    def get_place_schedule(self, *args, **kwargs):
        return self.place_schedule.get_schedule(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from six.moves.urllib import parse
from requests import RequestException
from school_api.client.api.base import BaseSchoolApi
from school_api.client.api.utils import get_alert_tip, get_view_state_from_html
from school_api.check_code import CHECK_CODE
from school_api.exceptions import IdentityException, CheckCodeException, LoginException, OtherException
from school_api.utils import to_binary, to_text


class Login(BaseSchoolApi):
    ''' 登录模块 '''

    def get_login(self, school, **kwargs):
        ''' 登录入口 与 异常处理 '''
        args = (school.login_url, school.exist_verify)
        try:
            res = self._get_api(*args, **kwargs)
        except OtherException as reqe:
            raise LoginException(self.code, to_text(str(reqe)))

        except RequestException:
            if school.proxies and not school.priority_proxy:
                # 使用内网代理
                if self._switch_proxy():
                    # 存在内网代理会话
                    return True
                try:
                    res = self._get_api(*args, **kwargs)
                except RequestException:
                    raise LoginException(self.code, '教务系统异常，使用代理登录失败')
            else:
                raise LoginException(self.code, '教务系统外网异常')

        self._handle_login_result(res, *args, **kwargs)

    def _get_login_payload(self, login_url, **kwargs):
        ''' 获取登录页面的 请求参数'''
        try:
            kwargs['timeout'] = 3
            res = self._get(login_url, **kwargs)
        except RequestException:

            # 首次请求可能出现 Connection aborted
            res = self._get(login_url, **kwargs)


        url_info = res.url.split(login_url)[0].split('/(')
        if len(url_info) == 2:
            self._update_url_token('/('+url_info[1])

        view_state = get_view_state_from_html(res.text)

        return {'__VIEWSTATE': view_state}

    def _get_api(self, login_url, exist_verify, **kwargs):
        # 登录请求
        code = ''
        login_types = ['学生', '教师', '部门']
        login_payload = self._get_login_payload(login_url, **kwargs)
        if exist_verify:
            res = self._get('/CheckCode.aspx')
            if res.content[:7] != to_binary('GIF89aH'):
                raise CheckCodeException(self.code, "验证码获取失败")
            code = CHECK_CODE.verify(res.content)

        account = self.account.encode('gb2312')
        payload = {
            'txtUserName': account,
            'TextBox1': account,
            'TextBox2': self.password,
            'TextBox3': code,
            'txtSecretCode': code,
            'RadioButtonList1': login_types[self.user_type].encode('gb2312'),
            'Button1': ' 登 录 '.encode('gb2312')
        }
        payload.update(login_payload)

        res = self._post(login_url, data=payload,
                         allow_redirects=False, **kwargs)
        return res

    def _handle_login_result(self, res, *args, **kwargs):
        # 登录成功之后，教务系统会返回 302 跳转
        if res.status_code == 500:
            raise LoginException(self.code, '教务系统请求异常')
        elif res.status_code != 302:
            tip = self._get_login_result_tip(res.text)
            if tip == '验证码不正确！！':
                # 首次验证码错误，则再次登录
                res = self._get_api(*args, **kwargs)
                if res.status_code != 302:
                    tip = self._get_login_result_tip(res.text)
                    if tip == '验证码不正确！！':
                        raise CheckCodeException(self.code, tip)
                else:
                    return True
            raise IdentityException(self.code, tip)

    def _get_login_result_tip(self, html):
        """ 获取获取html的弹框提示信息 """
        tip = get_alert_tip(html)
        if tip:
            return tip

        raise LoginException(self.code, '教务系统请求异常')

    def check_session(self):
        """ 检查登陆会话是否有效 """
        account = parse.quote(self.account.encode('gb2312'))
        try:
            res = self._head(self.school_url['HOME_URL'] + account)
            if res.status_code != 200:
                raise RequestException
        except RequestException:
            return False

        return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from school_api.exceptions import SchoolException, LoginException, PermissionException


class UserType():
    ''' 用户类型参数 '''
    STUDENT = 0
    TEACHER = 1
    DEPT = 2


class ScheduleType():
    ''' 课表类型参数 '''
    PERSON = 0
    CLASS = 1


class LoginFail():
    ''' 登录失败返回错误信息 '''

    def __init__(self, tip=''):
        self.tip = tip

    def __getattr__(self, name):
        def func(*args, **kwargs):
            return {'error': self.tip}
        return func

    def __nonzero__(self):
        return True


def error_handle(func):
    def wrapper(self, *args, **kwargs):
        if not self.school.use_ex_handle:
            result = func(self, *args, **kwargs)
        else:
            try:
                result = func(self, *args, **kwargs)

            except LoginException as reqe:
                result = LoginFail(reqe)

            except SchoolException as reqe:
                result = {'error': reqe}

        return result
    return wrapper


class ApiPermissions():
    ''' 接口权限判断 '''

    def __init__(self, permission_list):
        self.permission_list = permission_list

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            func_object = args[0]
            if func_object.user_type not in self.permission_list:
                raise PermissionException(func_object.school.code, '暂无该接口权限')
            return func(*args, **kwargs)
        return wrapper


def get_time_list(class_time):
    ''' 上课时间处理 '''
    time_list = {1: [], 2: [], 3: [], 4: []}
    time_text = "{} ~ {}"
    for index, times in enumerate(class_time):
        if index % 2 == 0:
            time_list[1].append(time_text.format(times[0], times[1]))
            time_list[2].append(time_text.format(times[0], class_time[index+1][1]))

            if index < 8:
                time_list[3].append(time_text.format(times[0], class_time[index+2][1]))
                time_list[4].append(time_text.format(times[0], class_time[index+3][1]))
    return time_list



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import inspect
import requests

from school_api.utils import to_text
from school_api.client.api.base import BaseSchoolApi
from school_api.client.api.utils import get_view_state_from_html
from school_api.client.login_management import LoginManagement


def _is_api_endpoint(obj):
    return isinstance(obj, BaseSchoolApi)


class BaseUserClient(LoginManagement):
    """docstring for BaseUserClient"""

    _proxy = None
    url_token = ''

    def __new__(cls, *args):
        self = super(BaseUserClient, cls).__new__(cls)
        api_endpoints = inspect.getmembers(self, _is_api_endpoint)
        for name, api in api_endpoints:
            api_cls = type(api)
            api = api_cls(self)
            setattr(self, name, api)
        return self

    def __init__(self, school, account, password, user_type):
        self._http = requests.Session()

        self.account = to_text(account)
        self.password = password
        self.user_type = user_type
        self.school = school.school
        self.base_url = school.base_url
        self.session = school.session

        self._http.headers.update({
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) '
                          'AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Chrome/62.0.3202.89 Safari/537.36',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Referer': self.base_url + self.school.login_url
        })
        if self.school.priority_proxy:
            self.switch_proxy()

    def _request(self, url_suffix, **kwargs):

        url = '{base}{url_token}{url_suffix}'.format(
            base=self.base_url,
            url_suffix=url_suffix,
            url_token=self.url_token
        )

        kwargs['timeout'] = kwargs.get('timeout', self.school.timeout)
        res = self._http.request(
            url=url,
            proxies=self._proxy,
            **kwargs
        )
        return res

    def get(self, url, **kwargs):
        return self._request(url, method='GET', **kwargs)

    def post(self, url, **kwargs):
        return self._request(url, method='POST', **kwargs)

    def head(self, url, **kwargs):
        return self._request(url, method='HEAD', **kwargs)

    def switch_proxy(self):
        """ 设置代理 """
        self.base_url = self.school.lan_url or self.base_url
        self._proxy = self.school.proxies
        self._http.headers.update({
            'Referer': self.base_url + self.school.login_url
        })

        if not self.school.priority_proxy:
            # 检查是否有登录会话
            return self.get_login_session()

        self.school.priority_proxy = True

    def get_view_state(self, url_suffix, **kwargs):
        """ 获取页面 view_state 值"""
        res = self.get(url_suffix, allow_redirects=False, **kwargs)
        res.raise_for_status()
        if res.status_code == 302:
            raise requests.TooManyRedirects

        return get_view_state_from_html(res.text)

    def update_url_token(self, url_token):
        # 兼容含token的教务系统请求地址 http://xxx.xxx/(35yxiq45pv0ojz45wcopgz45)/Default2.aspx
        self.url_token = url_token


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import time
from school_api.session import SessionStorage
from school_api.utils import to_text


class MemoryStorage(SessionStorage):
    ''' 非持久化缓存 不推荐使用'''

    def __init__(self, prefix=''):
        self._data = {}
        self.prefix = to_text(prefix)

    def key_name(self, key):
        return '{0}:{1}'.format(self.prefix, key)

    def get(self, key, default=None):
        key = self.key_name(key)
        value = None
        data = self._data.get(key, default)
        if data:
            if data['expires_at'] is True or data['expires_at'] > time.time():
                # 为True时：永不过期， 大于当前时间为：不到过期时间
                value = data['value']
            else:
                self.delete(key)
        return value

    def set(self, key, value, ttl=None):

        if value is None:
            return

        key = self.key_name(key)
        expires_at = not ttl or time.time() + ttl
        data = {'value': value, 'expires_at': expires_at}
        self._data[key] = data

    def delete(self, key):
        key = self.key_name(key)
        self._data.pop(key, None)

    def expires_time(self, key):
        data = self._data.get(self.key_name(key))
        return data['expires_at'] - time.time()
